/* This file is part of mailfrom filter.
   Copyright (C) 2005, 2006 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#include <sys/types.h>
#include <regex.h>
#include <ctype.h>
#ifdef HAVE_PTHREAD
# include <pthread.h>
#endif
#include <mailutils/mailutils.h>
#include <mfapi.h>
#include <xalloc.h>

/* Common error codes */
#ifdef HAVE_SYSEXITS_H
# include <sysexits.h>
#else
# define EX_OK          0       /* successful termination */
# define EX__BASE       64      /* base value for error messages */
# define EX_USAGE       64      /* command line usage error */
# define EX_DATAERR     65      /* data format error */
# define EX_NOINPUT     66      /* cannot open input */
# define EX_NOUSER      67      /* addressee unknown */
# define EX_NOHOST      68      /* host name unknown */
# define EX_UNAVAILABLE 69      /* service unavailable */
# define EX_SOFTWARE    70      /* internal software error */
# define EX_OSERR       71      /* system error (e.g., can't fork) */
# define EX_OSFILE      72      /* critical OS file missing */
# define EX_CANTCREAT   73      /* can't create (user) output file */
# define EX_IOERR       74      /* input/output error */
# define EX_TEMPFAIL    75      /* temp failure; user is invited to retry */
# define EX_PROTOCOL    76      /* remote error in protocol */
# define EX_NOPERM      77      /* permission denied */
# define EX_CONFIG      78      /* configuration error */
# define EX__MAX        78      /* maximum listed value */
#endif

/* Status values used throughout the program */
typedef enum {
	mf_success,
	mf_not_found,
	mf_failure,
	mf_temp_failure,

	mf_ston_conv,
        mf_divzero,
        mf_regcomp,
        mf_invip,
        mf_invcidr,
        mf_invtime,
        mf_dbfailure,
	mf_range,
	mf_url,
	mf_resolve,
	mf_ioerr,
	
	mf_status_count
} mf_status;

#define mf_resolved(c) ((c) == mf_success || (c) == mf_not_found)

#define EXMASK(n) ((unsigned)1<<(n))

/* SMTP (libmilter) states */
enum smtp_state {
	smtp_state_none,
	smtp_state_connect,
	smtp_state_first=smtp_state_connect,
	smtp_state_helo,
	smtp_state_envfrom,
	smtp_state_envrcpt,
	smtp_state_header,
	smtp_state_eoh,
	smtp_state_body,
	smtp_state_eom,

	smtp_state_count
};


/* The following defines are borrowed from intprops.h (gnulib) */

/* True if the arithmetic type T is signed.  */
#define TYPE_SIGNED(t) (! ((t) 0 < (t) -1))

/* Return zero if T can be determined to be an unsigned type.
   Otherwise, return 1.
   When compiling with GCC, INT_STRLEN_BOUND uses this macro to obtain a
   tighter bound.  Otherwise, it overestimates the true bound by one byte
   when applied to unsigned types of size 2, 4, 16, ... bytes.
   The symbol signed_type_or_expr__ is private to this header file.  */
#if __GNUC__ >= 2
# define signed_type_or_expr__(t) TYPE_SIGNED (__typeof__ (t))
#else
# define signed_type_or_expr__(t) 1
#endif

/* Bound on length of the string representing an integer type or expression T.
   Subtract 1 for the sign bit if T is signed; log10 (2.0) < 146/485;
   add 1 for integer division truncation; add 1 more for a minus sign
   if needed.  */
#define INT_STRLEN_BOUND(t) \
  ((sizeof (t) * CHAR_BIT - signed_type_or_expr__ (t)) * 146 / 485 \
   + signed_type_or_expr__ (t) + 1)

/* Bound on buffer size needed to represent an integer type or expression T,
   including the terminating null.  */
#define INT_BUFSIZE_BOUND(t) (INT_STRLEN_BOUND (t) + 1)

/* Size of the buffer able to accomodate mailfromd numeric type */
#define NUMERIC_BUFSIZE_BOUND INT_BUFSIZE_BOUND(long)


/* Libresolv-related defines and DNS functions */
#ifndef MAXPACKET
# define MAXPACKET 8192	/* max packet size used internally by BIND */
#endif 

#define MAXMXCOUNT 32
typedef char *mxbuf_t[MAXMXCOUNT];

mf_status resolve_ipstr(char *ipstr, char *hbuf, size_t hbsize);
mf_status getmx(char *ipstr, mxbuf_t mxbuf);
void freemx(mxbuf_t mxbuf);


/* Default file names */
#define DEFAULT_PIDFILE MAILFROMSTATEDIR "/mailfromd.pid"
#define DEFAULT_DATABASE MAILFROMSTATEDIR "/mailfromd.db"
#define DEFAULT_RATE_DATABASE MAILFROMSTATEDIR "/rates.db"
#define DEFAULT_GREYLIST_DATABASE MAILFROMSTATEDIR "/greylist.db"
#define DEFAULT_CONFIG_FILE SYSCONFDIR "/mailfromd.rc"


/* Debugging macros */

#define __DBG(lev) if (debug_level_p(__FILE__, lev))

#define debug(lev,text)                                                   \
 do                                                                       \
  __DBG(lev) {                                                            \
   if (source_info_option)                                                \
     debug_log("%s:%lu:%s: %s", __FILE__, __LINE__, __FUNCTION__, text);  \
   else                                                                   \
     debug_log("%s", text);                                               \
  }                                                                       \
 while (0)
#define debug1(lev,fmt,v)                                                 \
 do                                                                       \
  __DBG(lev) {                                                            \
   if (source_info_option)                                                \
    debug_log("%s:%lu:%s: " fmt, __FILE__, __LINE__, __FUNCTION__, v);    \
   else                                                                   \
    debug_log(fmt, v);                                                    \
  }                                                                       \
 while (0)
#define debug2(lev,fmt,a,b)                                               \
 do                                                                       \
  __DBG(lev) {                                                            \
   if (source_info_option)                                                \
     debug_log("%s:%lu:%s: " fmt, __FILE__, __LINE__, __FUNCTION__, a, b);\
   else                                                                   \
     debug_log(fmt, a, b);                                                \
  }                                                                       \
 while (0)
#define debug3(lev,fmt,a,b,c)                                             \
 do                                                                       \
  __DBG(lev) {                                                            \
   if (source_info_option)                                                \
    debug_log("%s:%lu:%s: " fmt, __FILE__, __LINE__, __FUNCTION__, a, b, c);\
   else                                                                   \
    debug_log(fmt, a, b, c);                                              \
  }                                                                       \
 while (0)
#define debug4(lev,fmt,a,b,c,d)                                           \
 do                                                                       \
  __DBG(lev) {                                                            \
   if (source_info_option)                                                \
    debug_log("%s:%lu:%s: " fmt, __FILE__, __LINE__, __FUNCTION__,        \
              a, b, c, d);                                                \
   else                                                                   \
    debug_log(fmt, a, b, c, d);                                           \
  }                                                                       \
 while (0)
#define debug5(lev,fmt,a,b,c,d,e)                                         \
 do                                                                       \
  __DBG(lev) {                                                            \
   if (source_info_option)                                                \
    debug_log("%s:%lu:%s: " fmt,                                          \
              __FILE__, __LINE__, __FUNCTION__, a, b, c, d, e);           \
  else                                                                    \
    debug_log(fmt, a, b, c, d, e);                                        \
  }                                                                       \
 while (0)
#define debug6(lev,fmt,a,b,c,d,e,f)                                       \
 do                                                                       \
  __DBG(lev) {                                                            \
   if (source_info_option)                                                \
    debug_log("%s:%lu:%s: " fmt,                                          \
              __FILE__, __LINE__, __FUNCTION__, a, b, c, d, e, f);        \
   else                                                                   \
    debug_log(fmt, a, b, c, d, e, f);                                     \
  }                                                                       \
 while (0)

void debug_log(char *fmt, ...);
void enable_module_trace(char *name);
void disable_module_trace(char *name);
void enable_prog_trace(const char *modlist);
void disable_prog_trace(const char *modlist);
int debug_level_p(const char *file, int level);
void enable_debug(const char *file);
void disable_debug(const char *file);
void enable_debug_list(const char *spec);
void disable_debug_list(const char *spec);


/* Global variables declarations */
#define MAILFROMD_DAEMON        0
#define MAILFROMD_TEST          1
#define MAILFROMD_DELETE        2
#define MAILFROMD_LIST          3
#define MAILFROMD_EXPIRE        4
#define MAILFROMD_COMPACT       5    
#define MAILFROMD_SHOW_DEFAULTS 6

extern int mode;
extern int debug_level;
extern unsigned optimization_level;
extern enum smtp_state test_state;
extern time_t negative_expire_interval;
extern int predict_next_option;
extern double predict_rate;
extern char *smtp_domain;
extern char *postmaster_email;
extern int regex_flags;
extern size_t stack_size;
extern size_t variable_count;
extern int ignore_failed_reads_option;
extern int source_info_option;
extern int do_transcript;
extern char *portspec;
extern unsigned long source_address;
extern char *pidfile;
extern time_t io_timeout;
extern time_t connect_timeout;
extern time_t response_timeout;
extern int foreground;
extern int force_remove;
extern int config_dump_macros;
extern int config_dump_xref;
extern size_t lock_retry_count_option;
extern time_t lock_retry_timeout_option;


/* Configuration file parser */

typedef enum {
	dtype_unspecified,
	dtype_string,
	dtype_number
} data_type_t;

/* Parse tree node */
typedef struct node NODE;

/* Input location */
struct locus {
	char *file;         /* File name */
	size_t line;        /* Line number */
};

struct if_node {            /* if statement */
	NODE *cond;         /* Condition */
	NODE *if_true;      /* True branch */
	NODE *if_false;     /* False branch */
};

struct poll_data {                 /* Poll test */
	NODE *email;               
	NODE *client_addr;
	NODE *ehlo;
	NODE *mailfrom;
};

/* Opcodes for available binary operations */
enum bin_opcode {
	bin_and,
	bin_or,
	bin_eq,
	bin_ne,
	bin_lt,
	bin_le,
	bin_gt,
	bin_ge,
	bin_match,
	bin_fnmatch,
	bin_add,
	bin_sub,
	bin_mul,
	bin_div
};

#define QUALIFIER_MX      0x1
#define QUALIFIER_REGFREE 0x2

struct bin_node {                 /* A binary operation */
	enum bin_opcode opcode;   /* Operation code */
	int qualifier;            /* Used to indicate MX MATCHES/FNMATCHES */
	NODE *arg[2];             /* arg[0] - left side argument,
				     arg[1] - right side argument */
};

/* Unary operations - currently only one */
enum un_opcode {
	unary_not,
	unary_minus
};

struct un_node {                  /* A unary operation node */ 
	enum un_opcode opcode;    /* Operation code */
	data_type_t result_type;  /* Result type */
	NODE *arg;                /* Argument */ 
};

/* Return action node: accept/reject/tempfail/continue */
struct return_node {              
	sfsistat stat;            /* Return status */
	char *code;               /* Code */
	char *xcode;              /* Extended code */
	NODE *message;            /* Subtree producing the textual message */
};

enum header_opcode {              /* Operation on a header */
	header_add,               /* Add a header */
	header_replace,           /* Replace a header value */
	header_delete             /* Delete a header */ 
};

struct old_header_node {              
	enum header_opcode opcode; /* Operation code */
	char *name;               /* Header name */
	char *value;              /* Header value */
};

struct header_node {              
	enum header_opcode opcode; /* Operation code */
	char *name;               /* Header name */
	NODE *value;              /* Header value */
};

struct rate_node {                /* Rate test */
	NODE *email;              /* Originator email */
	double limit;             /* Sending rate limit */
};

struct builtin_node {             /* Call to a built-in function */
	const struct builtin *builtin; /* Buit-in function */ 
	NODE *args;                    /* Actual arguments */
};

struct concat_node {              /* Concatenation of arg[0] and arg[1] */ 
	NODE *arg[2];
};

struct asgn_node {                /* Assignement */ 
	struct variable *var;     /* Variable */ 
	NODE *node;               /* Expression to be evaluated and assigned
				     to it */ 
};

struct catch_node {
	int count;
	unsigned exmask;
	NODE *node;
};

struct function_call {
	struct function *func;
	NODE *args;
};

struct value {
	data_type_t type;
	union {
		long number;
		char *string;
	} v;
};

struct valist {
	struct valist *next;
	struct value value;
};

struct case_stmt {
	struct case_stmt *next;
	struct locus locus;
	struct valist *valist;
	NODE *node;
};

struct switch_stmt {
	NODE *node;
	struct case_stmt *cases;
};

struct cast_node {
	data_type_t data_type;
	NODE *node;
};

struct positional_arg {
	data_type_t data_type;
	int number;
};

/* Types of parse tree nodes */
enum node_type {
	node_type_string,
	node_type_symbol,
	node_type_number,
	node_type_if,
	node_type_bin,
	node_type_un,
	node_type_next,
	node_type_result,
	node_type_header,
	node_type_builtin,
	node_type_concat,
	node_type_variable,
	node_type_asgn,
	node_type_arg,
	node_type_push,
	node_type_regcomp,
	node_type_catch,
	node_type_echo,
	node_type_return,
	node_type_call,
	node_type_switch,
	node_type_backref,
	node_type_cast
};

/* Parse tree node */
struct node {
	NODE *next;                   /* Link to the next node */
	enum node_type type;          /* Node type */
	struct locus locus;           /* Corresponding input location */
	union {
		char *string;
		long number;
		struct if_node cond;
		struct bin_node bin;
		struct un_node un;
		struct return_node ret;
		struct header_node hdr;
		struct rate_node rate;
		struct builtin_node builtin;
		NODE *node;
		struct concat_node concat;
		struct variable *variable;
		struct asgn_node asgn;
		void *val;
		struct catch_node catch;
		struct function_call call;
		struct switch_stmt switch_stmt;
		struct cast_node cast;
		struct positional_arg arg;
	} v;
};


/* Expressions, built-in functions and variables */

typedef struct eval_environ *eval_environ_t;  /* Evaluation environment */
typedef unsigned long prog_counter_t;         /* Program counter */  

/* Data types */

struct builtin {
	char *name;                       /* Built-in function name */
	void (*handler) (eval_environ_t); /* C handler */
	size_t parmcount;                 /* Number of parameters */
	data_type_t *parmtype;            /* Parameter types */
	size_t optcount;                  /* Number of optional parameters */
	data_type_t rettype;              /* Return type */
	unsigned statemask;               /* States where the function can
					     be used */
	int capture;                      /* Does it need message capturing */
};

struct function {      /* User-defined function */
	char *name;    /* Function name */
	struct locus locus;      /* Location of the definition */
	NODE *node;              /* Function definition (syntax tree) */
	prog_counter_t entry;    /* Entry point to the function code */
	unsigned exmask;         /* Exception mask */
	size_t parmcount;        /* Number of arguments */
	data_type_t *parmtype;   /* Argument types */
	data_type_t rettype;     /* Return type */
};	

typedef enum {
	storage_extern,
	storage_auto,
	storage_param
} storage_class_t;

#define VAR_VOLATILE  0x1   /* Variable can change */
#define VAR_EXTERN    0x2   /* Variable is external (for future use) */

struct variable {
	char *name;                /* Variable name */
	data_type_t type;          /* Variable type */
	storage_class_t storage_class; /* Storage class */
	size_t off;                /* Offset in the data segment */
	struct locus locus;        /* Location of the definition */
	struct variable *shadowed; /* Points to the variable shadowed by
				      this one */
	unsigned flags;            /* VAR_* flags */
	mu_list_t xref;            /* List of struct locus */
};

#define SYM_UNDEF     0      /* The entry is not defined */
#define SYM_BUILTIN   1      /* The entry holds a builtin function */
#define SYM_FUNC      2      /* The entry holds a user-defined function */
#define SYM_VARIABLE  3      /* The entry holds a variable */
#define SYM_DB_FORMAT 4      /* Database format handler */

typedef int (*symbol_enumerator_t)(void *sym, void *data);
int symbol_enumerate(int state, symbol_enumerator_t fun, void *data);
	
void builtin_setup(void);
void builtin_post_setup(void);
void print_code(void);
void va_builtin_install (char *name, void (*handler) (eval_environ_t),
			 data_type_t rettype, size_t argcount, ...);
void va_builtin_install_ex (char *name, void (*handler) (eval_environ_t),
			    unsigned statemsk, int capture,
			    data_type_t rettype, size_t argcount,
			    size_t optcount, ...);

const struct builtin *builtin_lookup(const char *name);
struct variable *variable_install(const char *name);
struct variable *variable_lookup(const char *name);
void variable_remove(struct variable *var);
struct variable *builtin_variable_install(const char *name,
					  data_type_t type,
					  unsigned flags);
struct function *function_install(char *name, int parmcnt,
				  data_type_t *parmtypes,
				  data_type_t rettype,
				  const struct locus *locus);
struct function *function_lookup(const char *name);


/* Parser functions */
int yyparse();
int yylex();
int yyerror(char *s);
int source(char *, int);
void onblock(int enable);
int parse_config(char *name, int ydebug, int ldebug);
void parse_pragma(char *text);
const struct locus *get_locus(void);
const char *header_command_str(enum header_opcode opcode);
const char *sfsistat_str(sfsistat stat);
const char *mf_status_str(mf_status stat);
const char *storage_class_str(storage_class_t sc);
int string_to_stat(char *str, mf_status *status);
enum smtp_state string_to_state(char *name);
const char *state_to_string(enum smtp_state state);
void parse_error(char *fmt, ...);
void parse_error_locus(const struct locus *loc, char *fmt, ...);
int parse_time_interval(char *str, time_t *pint, char **endp);
void print_config(void);
void print_config_macros(void);
const char *function_name(void);

char *string_alloc(char *str, size_t len);
void string_begin(void);
void string_add(char *str, size_t len);
void string_add_char(unsigned char c);
char *string_finish(void);


/* Data types and declarations for handling compiled configuration code */

typedef void (*instr_t) (eval_environ_t env); /* Instruction */

/* Entry points for each SMTP state */
extern prog_counter_t entry_point[smtp_state_count];

/* Available instructions */
void instr_xchg(eval_environ_t env);
void instr_ston(eval_environ_t env);
void instr_ntos(eval_environ_t env);
void instr_locus(eval_environ_t env);
void instr_push_immediate(eval_environ_t env);
void instr_stkalloc(eval_environ_t env);
void instr_symbol(eval_environ_t env);
void instr_bz(eval_environ_t env);
void instr_bnz(eval_environ_t env);
void instr_jmp(eval_environ_t env);
void instr_dup(eval_environ_t env);
void instr_cmp(eval_environ_t env);
void instr_pop(eval_environ_t env);
void instr_regex(eval_environ_t env);
void instr_fnmatch_mx(eval_environ_t env);
void instr_regex_mx(eval_environ_t env);
void instr_regcomp(eval_environ_t env);
void instr_fnmatch(eval_environ_t env);
void instr_not(eval_environ_t env);
void instr_next(eval_environ_t env);
void instr_result(eval_environ_t env);
void instr_header(eval_environ_t env);
void instr_eqn(eval_environ_t env);
void instr_eqs(eval_environ_t env);
void instr_nen(eval_environ_t env);
void instr_nes(eval_environ_t env);
void instr_ltn(eval_environ_t env);
void instr_lts(eval_environ_t env);
void instr_len(eval_environ_t env);
void instr_les(eval_environ_t env);
void instr_gtn(eval_environ_t env);
void instr_gts(eval_environ_t env);
void instr_gen(eval_environ_t env);
void instr_ges(eval_environ_t env);
void instr_builtin(eval_environ_t env);
void instr_concat(eval_environ_t env);
void instr_variable(eval_environ_t env);
void instr_asgn(eval_environ_t env);
void instr_svalidate(eval_environ_t env);
void instr_arg(eval_environ_t env);
void instr_add(eval_environ_t env);
void instr_sub(eval_environ_t env);
void instr_mul(eval_environ_t env);
void instr_div(eval_environ_t env);
void instr_neg(eval_environ_t env);
void instr_catch(eval_environ_t env);
void instr_echo(eval_environ_t env);
void instr_return(eval_environ_t env);
void instr_retcatch(eval_environ_t env);
void instr_adjust(eval_environ_t env);
void instr_popreg(eval_environ_t env);
void instr_pushreg(eval_environ_t env);
void instr_funcall(eval_environ_t env);
void instr_xlat(eval_environ_t env);
void instr_xlats(eval_environ_t env);
void instr_jreg(eval_environ_t env);
void instr_saveex(eval_environ_t env);
void instr_restex(eval_environ_t env);
void instr_backref(eval_environ_t env);
void instr_auto(eval_environ_t env);

/* Functions for accessing %rcpt_count */
unsigned long get_rcpt_count(eval_environ_t env);
void clear_rcpt_count(eval_environ_t env);
void incr_rcpt_count(eval_environ_t env);
void set_last_poll_result(eval_environ_t env,
			  const char *host_addr, const char *send,
			  const char *recv);
void set_cache_used(eval_environ_t env, int value);

/* Code generation and diagnostics functions */
void code_put(prog_counter_t pos, void *ptr);
instr_t code_peek(prog_counter_t pos);
prog_counter_t code_reserve(size_t count);

prog_counter_t code_immediate(const void *ptr);
prog_counter_t code_instr(const instr_t ptr);
prog_counter_t code_get_counter();
void dump_code(prog_counter_t start, prog_counter_t end);

/* Evaluation environment functions */
eval_environ_t
create_environment(SMFICTX *ctx,
		   char *(*getsym)(void *data, char *str),
		   int (*setreply)(void *data, char *code, char *xcode,
				   char *message),
		   void (*setheader)(void *data, struct old_header_node *hdr),
		   void *data);
void destroy_environment(eval_environ_t env);
void env_init(eval_environ_t env);
int env_catch(eval_environ_t env, mf_status status);
void env_push_string(eval_environ_t env, char *arg);
void env_push_number(eval_environ_t env, long arg);
void env_make_frame(eval_environ_t env);
void env_leave_frame(eval_environ_t env, int nargs);
int env_set_variable(eval_environ_t env, char *ident, char *value);
int eval_environment(eval_environ_t env, prog_counter_t pc);
sfsistat environment_get_status(eval_environ_t env);
const char *environment_get_null_symbol(eval_environ_t env,
					struct locus *locus);
SMFICTX *env_get_context(eval_environ_t env);
mu_stream_t env_get_stream(eval_environ_t env);
int env_capture_start(eval_environ_t env);
int env_capture_write(eval_environ_t env, const char *buf, size_t size);
int env_capture_write_args(eval_environ_t env, ...);
void env_final_gc(eval_environ_t env);


/* Runtime functions */
const char *mailfromd_msgid(SMFICTX *ctx);
char *mailfromd_timestr(time_t timestamp, char *timebuf, size_t bufsize);
void logmsg(int prio, const char *fmt, ...);
void log_status(sfsistat status, SMFICTX *ctx);
void trace(const char *fmt, ...);
int convert_rate(const char *arg, double *rate);
extern void priv_setup(void);
void mailfromd_daemon(void);

/* DNS dictionary */
void dict_init(mu_list_t *dict);
char *dict_install(mu_list_t dict, char *name, char *value);
void dict_destroy(mu_list_t *dict);
char *dict_getsym(void *data, char *str);

void mailfromd_test(int argc, char **argv);

int set_option(char *name, char *value, int override);
int sendmail_mlreply(SMFICTX * ctx, const char *rcode, const char *xcode,
		     char *message);
void milter_enable_state(enum smtp_state state);
int xeval(eval_environ_t env, enum smtp_state tag);

mf_status listens_on (const char *client_addr, int port);
mf_status check_on_host(eval_environ_t env, char *email, char *client_addr,
			char *ehlo, char *mailfrom);
mf_status check_mx_records(eval_environ_t env, char *email, char *client_addr,
			   char *ehlo, char *mailfrom, int *pcount);
mf_status method_strict(eval_environ_t env, char *email, char *client_addr,
			char *ehlo, char *mailfrom);
mf_status method_standard(eval_environ_t env, char *email, char *ehlo,
			  char *mailfrom);
int relayed_domain_p(char *name);
int compare_string(const void *item, const void *value);


/* DBM support */
#include "mu_dbm.h"

typedef void (*db_item_printer_t)(const char *key, size_t size,
				 const void *content);
typedef int (*db_expire_t)(const void *content);

struct db_format {
	char *name;
	char *dbname;
	time_t expire_interval;
	db_item_printer_t print_item;
	db_expire_t expire;
};

struct db_format *db_format_install(struct db_format *fmt);
struct db_format *db_format_lookup(const char *name);

#ifdef USE_DBM
# define cache_get _mailfrom_cache_get
# define cache_insert _mailfrom_cache_insert
# define cache_get2 _mailfrom_cache_get2
# define cache_insert2 _mailfrom_cache_insert2

mf_status cache_get(char *email);
void cache_insert(char *email, mf_status rc);
mf_status cache_get2(char *email, char *client_addr);
void cache_insert2(char *email, char *client_addr, mf_status rc);

mf_status get_rate(char *email, double *ret);

int db_list_item(char *dbname, char *key, db_item_printer_t fun);
int db_list(char *dbname, db_item_printer_t fun);
int db_expire(char *dbname, db_expire_t fun);
int db_delete(char *dbname, char *id);
int db_compact(char *dbname, db_expire_t fun);

extern struct db_format *cache_format;
extern struct db_format *rate_format;
extern struct db_format *greylist_format;
extern char *time_format_string;

size_t format_time_str(FILE *fp, time_t timestamp);

#else
# define cache_get(email) mf_failure
# define cache_insert(email, rc)
# define cache_get2(email, client_addr) mf_failure
# define cache_insert2(email, client_addr, rc)
# define get_rate(email, ret) mf_failure

#define cache_format    NULL
#define rate_format     NULL
#define greylist_format NULL

#define db_list_item(dbname, key, fun) 1
#define db_list(dbname, fun) 1
#define db_expire(dbname, fun) 1
#define db_delete(dbname, id) 1
#define db_compact(dbname, fun) 1
#endif

#ifdef HAVE_PTHREAD
# define MUTEX_DCL(name) \
 static pthread_mutex_t name = PTHREAD_MUTEX_INITIALIZER;
# define MUTEX_LOCK(name) \
 pthread_mutex_lock(&name)
# define MUTEX_UNLOCK(name) \
 pthread_mutex_unlock(&name);
#else
# define MUTEX_DCL(name)
# define MUTEX_LOCK(name)
# define MUTEX_UNLOCK(name)
#endif

#if !defined res_ninit && !defined HAVE_RES_NINIT
# define res_ninit(s)
#endif

#if !defined res_nclose && !defined HAVE_RES_NCLOSE
# define res_nclose(s)
#endif

#if !defined res_nquery && !defined HAVE_RES_NQUERY
# define res_nquery(stat,name,class,type,answer,anslen) \
 res_query(name,class,type,answer,anslen)
#endif

#if !defined res_nquerydomain && !defined HAVE_RES_NQUERYDOMAIN
# define res_nquerydomain(stat,name,domain,class,type,answer,anslen) \
 res_querydomain(name,domain,class,type,answer,anslen)
#endif



